/**
 * SmartCGMS - continuous glucose monitoring and controlling framework
 * https://diabetes.zcu.cz/
 *
 * Copyright (c) since 2018 University of West Bohemia.
 *
 * Contact:
 * diabetes@mail.kiv.zcu.cz
 * Medical Informatics, Department of Computer Science and Engineering
 * Faculty of Applied Sciences, University of West Bohemia
 * Univerzitni 8, 301 00 Pilsen
 * Czech Republic
 * 
 * 
 * Purpose of this software:
 * This software is intended to demonstrate work of the diabetes.zcu.cz research
 * group to other scientists, to complement our published papers. It is strictly
 * prohibited to use this software for diagnosis or treatment of any medical condition,
 * without obtaining all required approvals from respective regulatory bodies.
 *
 * Especially, a diabetic patient is warned that unauthorized use of this software
 * may result into severe injure, including death.
 *
 *
 * Licensing terms:
 * Unless required by applicable law or agreed to in writing, software
 * distributed under these license terms is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *
 * a) This file is available under the Apache License, Version 2.0.
 * b) When publishing any derivative work or results obtained using this software, you agree to cite the following paper:
 *    Tomas Koutny and Martin Ubl, "SmartCGMS as a Testbed for a Blood-Glucose Level Prediction and/or 
 *    Control Challenge with (an FDA-Accepted) Diabetic Patient Simulation", Procedia Computer Science,  
 *    Volume 177, pp. 354-362, 2020
 */

#include "generated_model.h"

#include <scgms/rtl/SolverLib.h>

// this selects GCTv3 support implementations
using namespace gct3_model;

/**********************************
 * Generated model implementation *
 **********************************/

CGenerated123_Model::CGenerated123_Model(scgms::IModel_Parameter_Vector* parameters, scgms::IFilter* output) :
	CBase_Filter(output),
	// BP NOTE: jen generujte spravne jmeno namespace
	mParameters(scgms::Convert_Parameters<generated123_model::TParameters>(parameters, generated123_model::default_parameters.vector))
{
	mCompartments.resize(Compartment_Vector_Size);

	// BP NOTE: pro kazdy kompartment generujte depot dle sablony
	auto& c0 =  mCompartments[0].Create_Depot(0.0, false);
	auto& c1 = mCompartments[1].Create_Depot(0.0, false);
	auto& c2 = mCompartments[2].Create_Depot(0.0, false);
	auto& c3 = mCompartments[3].Create_Depot(0.0, false);

	// BP NOTE: nechte jak je
	auto& universal_source  = mCompartments[Universal_Source_Idx].Create_Depot<CSource_Depot>(1.0, false);
	auto& universal_sink = mCompartments[Universal_Sink_Idx].Create_Depot<CSink_Depot>(0.0, false);

	universal_source.Set_Persistent(true);
	universal_source.Set_Name(L"Source");

	universal_sink.Set_Persistent(true);
	universal_sink.Set_Name(L"Sink");

	// BP NOTE: kazdy depot kompartmentu nastavte na perzistentni a pojmenujte jej
	c0.Set_Persistent(true);
	c0.Set_Name(L"C0");

	c1.Set_Persistent(true);
	c1.Set_Name(L"C1");

	c2.Set_Persistent(true);
	c2.Set_Name(L"C2");

	c3.Set_Persistent(true);
	c3.Set_Name(L"C3");

	// BP NOTE: ted budeme specifikovat vazby

/*

Priklad zde:

A) Kompartment 0 je spojeny s kompartmentem 1 exponencialnim prenosem s parametrem 2 (C0 --(p2)--> C1)
B) Kompartment 0 je spojeny s kompartmentem 2 exponencialnim prenosem s parametrem 3 (C0 --(p3)--> C2)
C) Kompartment 1 je spojeny s kompartmentem 3 difuznim prenosem s parametrem 0 (C1 <--(p0)--> C3)            -- pozn. difuzni vazba = mnozstvi odchazi z kompartmentu, kde je ho vic do toho, kde je ho mene
D) Z kompartmentu 2 dochazi k exponencialni eliminaci s parametrem 1 (C2 --(p1)--> _)
E) V kompartmentu 0 dochazi ke konstantni appearance s parametrem 4 (_ --(p4)--> C0)

*/

	// A)
	c0.Link_To<CConstant_Unbounded_Transfer_Function>(
		c1,
		CTransfer_Function::Start,
		CTransfer_Function::Unlimited,
		mParameters.vector[2]
	);

	// B)
	c0.Link_To<CConstant_Unbounded_Transfer_Function>(
		c2,
		CTransfer_Function::Start,
		CTransfer_Function::Unlimited,
		mParameters.vector[3]
	);

	// C)
	c1.Link_To<CTwo_Way_Diffusion_Unbounded_Transfer_Function>(
		c3,
		CTransfer_Function::Start,
		CTransfer_Function::Unlimited,
		std::ref(c1), std::ref(c3),		// BP NOTE: pozor, u difuze uvadime reference na kompartmenty jeste i zde
		mParameters.vector[0]
	);

	// D)
	c2.Link_To<CConstant_Unbounded_Transfer_Function>(universal_sink,
		CTransfer_Function::Start,
		CTransfer_Function::Unlimited,
		mParameters.vector[1]);

	// E)
	universal_source.Link_To<CConstant_Unbounded_Transfer_Function>(c0,
		CTransfer_Function::Start,
		CTransfer_Function::Unlimited,
		mParameters.vector[4]);
}

void CGenerated123_Model::Emit_All_Signals(double time_advance_delta) {

	const double _T = mLast_Time;

	// BP NOTE: tohle generujte podle toho, jaky kompartment uzivatel oznacil jako vystupni - parujte signal s kompartmentem
	// tady na prikladu: vystupni signal 1 je generovat z kompartmentu s indexem 2
	Emit_Signal_Level(generated123_model::signal_Output_1, _T, mCompartments[2].Get_Quantity());
}

HRESULT CGenerated123_Model::Do_Execute(scgms::UDevice_Event event) {
	HRESULT res = S_FALSE;

	if (!std::isnan(mLast_Time)) {

		if (event.event_code() == scgms::NDevice_Event_Code::Level && event.device_time() >= mLast_Time) {

			if (event.signal_id() == generated123_model::signal_Input_1) {

				// BP NOTE: jestli uzivatel zadal vstupni signal, musel jej navazat na kompartment
				// tohle propojuje kompartment se vstupnim signalem

				// Na prikladu: vstupni signal 1 (viz podminka nahore) je propojen s kompartmentem 3

				CDepot& depot = mCompartments[Universal_Source_Idx].Create_Depot(event.level(), false);
				depot.Set_Name(std::wstring(L"External_Source"));

				depot.Link_To<CConstant_Bounded_Transfer_Function>(
					mCompartments[3].Get_Persistent_Depot(), // BP NOTE: tady se specifikuje kompartment
					event.device_time(),
					5 * scgms::One_Minute,
					1.0
				);

				res = S_OK;
			}
		}
	}

	if (res == S_FALSE)
		res = mOutput.Send(event);

	return res;
}

// BP NOTE: nechte jak je
HRESULT CGenerated123_Model::Do_Configure(scgms::SFilter_Configuration configuration, refcnt::Swstr_list& error_description) {
	return S_OK;
}

// BP NOTE: nechte jak je
HRESULT IfaceCalling CGenerated123_Model::Step(const double time_advance_delta) {

	HRESULT rc = E_FAIL;
	if (time_advance_delta > 0.0) {

		constexpr size_t microStepCount = 5;

		const double microStepSize = time_advance_delta / static_cast<double>(microStepCount);
		const double oldTime = mLast_Time;
		const double futureTime = mLast_Time + time_advance_delta;

		// stepping scope
		{
			for (size_t i = 0; i < microStepCount; i++) {

				// step all compartments
				std::for_each(std::execution::par_unseq, mCompartments.begin(), mCompartments.end(), [this](CCompartment& comp) {
					comp.Step(mLast_Time);
				});

				// commit all compartments
				std::for_each(std::execution::par_unseq, mCompartments.begin(), mCompartments.end(), [this](CCompartment& comp) {
					comp.Commit(mLast_Time);
				});

				mLast_Time = oldTime + static_cast<double>(i) * microStepSize;
			}
		}

		mLast_Time = futureTime;

		Emit_All_Signals(time_advance_delta);

		rc = S_OK;
	}
	else if (time_advance_delta == 0.0) {
		Emit_All_Signals(time_advance_delta);
		rc = S_OK;
	}

	return rc;
}

// BP NOTE: nechte jak je, jen vlozte spravne jmeno namespace
HRESULT CGenerated123_Model::Emit_Signal_Level(const GUID& signal_id, double device_time, double level) {

	scgms::UDevice_Event evt{ scgms::NDevice_Event_Code::Level };

	evt.device_id() = generated123_model::model_id; // BP NOTE: tady
	evt.device_time() = device_time;
	evt.level() = level;
	evt.signal_id() = signal_id;
	evt.segment_id() = mSegment_Id;

	return mOutput.Send(evt);
}

// BP NOTE: nechte jak je
HRESULT IfaceCalling CGenerated123_Model::Initialize(const double current_time, const uint64_t segment_id) {

	if (!std::isnan(mLast_Time)) {
		return E_ILLEGAL_STATE_CHANGE;
	}

	mLast_Time = current_time;
	mSegment_Id = segment_id;

	for (auto& cmp : mCompartments) {
		cmp.Init(current_time);
	}

	return S_OK;
}
